searchNodes=[{"doc":"Reusing the description from the paper: Pattern synonyms allow to abstract over patterns used in pattern matching, notably by allowing to use computation instead of being limited to concrete data. Pattern metonyms are an implementations of this, but because of the limitation of the language, it can not be considered the same, so metonyms was chosen as a synonym to synonyms. Unlike in Haskell, few metonym definitions can be used with case/2 , other forms of patterns can only be used in combination with view/2 .","ref":"PatternMetonyms.html","title":"PatternMetonyms","type":"module"},{"doc":"view with named functions use PatternMetonyms defv id ( ( Function . identity ( ) -&gt; x ) ) , do : x","ref":"PatternMetonyms.html#defv/2","title":"PatternMetonyms.defv/2","type":"macro"},{"doc":"View with anonymous functions import PatternMetonyms id = fnv do ( Function . identity ( ) -&gt; x ) -&gt; x end","ref":"PatternMetonyms.html#fnv/1","title":"PatternMetonyms.fnv/1","type":"macro"},{"doc":"Macro used to define pattern metonyms. There are three types of pattern, listed below with examples: Implicitly Bidirectional The simplest type of pattern, because of its symmetry requirement, it can only be defined using concrete data (or adapted macro). Therefore no computation is allowed, and they are thus compatible with case and function heads. They take the form: pattern &lt; name &gt; ( &lt; [ variables ] &gt; ) = &lt; pattern &gt; where pattern reuses the variables iex&gt; defmodule DoctestTPX do ...&gt; import PatternMetonyms ...&gt; ...&gt; pattern ok ( x ) = { :ok , x } ...&gt; pattern error ( x ) = { :error , x } ...&gt; ...&gt; pattern cons ( x , xs ) = [ x | xs ] ...&gt; ...&gt; def foo ( x ) do ...&gt; view x do ...&gt; ok ( a ) -&gt; a ...&gt; error ( b ) -&gt; b ...&gt; end ...&gt; end ...&gt; ...&gt; def bar ( x ) do ...&gt; case x do ...&gt; ok ( a ) -&gt; a ...&gt; error ( b ) -&gt; b ...&gt; end ...&gt; end ...&gt; ...&gt; def baz ( ok ( a ) ) , do : a ...&gt; def baz ( error ( b ) ) , do : b ...&gt; ...&gt; def mk_ok ( x ) , do : ok ( x ) ...&gt; ...&gt; def blorg ( xs ) do ...&gt; view xs do ...&gt; cons ( x , xs ) -&gt; cons ( x , Enum . map ( xs , fn x -&gt; - x end ) ) ...&gt; end ...&gt; end ...&gt; end iex&gt; DoctestTPX . foo ( { :ok , :banana } ) :banana iex&gt; DoctestTPX . foo ( { :error , :split } ) :split iex&gt; DoctestTPX . bar ( { :ok , :peach } ) :peach iex&gt; DoctestTPX . baz ( { :error , :melba } ) :melba iex&gt; DoctestTPX . mk_ok ( :melba ) { :ok , :melba } iex&gt; DoctestTPX . blorg ( [ 1 , 2 , 3 ] ) [ 1 , - 2 , - 3 ] Unidirectional This type of pattern is read only, it may be used as abstraction over pattern matching on concrete data type that can not be reused to construct data, or as abstraction over views, as explained in view/2 . They take the form: pattern &lt; name &gt; ( &lt; [ variables ] &gt; ) &lt;- &lt; pattern &gt; pattern &lt; name &gt; ( &lt; [ variables ] &gt; ) &lt;- ( &lt; function &gt; -&gt; &lt; pattern &gt; ) where pattern reuses the variables . (function -&gt; pattern) is called a view iex&gt; defmodule DoctestTPY do ...&gt; import PatternMetonyms ...&gt; ...&gt; pattern head ( x ) &lt;- [ x | _ ] ...&gt; ...&gt; pattern rev_head ( x ) &lt;- ( reverse ( ) -&gt; head ( x ) ) ...&gt; ...&gt; def reverse ( xs ) , do : Enum . reverse ( xs ) ...&gt; ...&gt; def foo ( xs ) do ...&gt; view xs do ...&gt; head ( x ) -&gt; x ...&gt; [ ] -&gt; [ ] ...&gt; end ...&gt; end ...&gt; ...&gt; def bar ( x ) do ...&gt; case x do ...&gt; head ( a ) -&gt; a ...&gt; [ ] -&gt; [ ] ...&gt; end ...&gt; end ...&gt; ...&gt; def baz ( head ( a ) ) , do : a ...&gt; ...&gt; def blorg ( xs ) do ...&gt; view xs do ...&gt; rev_head ( x ) -&gt; x ...&gt; end ...&gt; end ...&gt; end iex&gt; DoctestTPY . foo ( [ 1 , 2 , 3 ] ) 1 iex&gt; DoctestTPY . bar ( [ 1 , 2 , 3 ] ) 1 iex&gt; DoctestTPY . baz ( [ 1 , 2 , 3 ] ) 1 iex&gt; DoctestTPY . blorg ( [ 1 , 2 , 3 ] ) 3 Explicitly bidirectional This type of pattern allows the same kind of abstraction as unidirectional one, but also permit defining how to construct data from computation (if necessary). They take the form: pattern ( &lt; name &gt; ( &lt; [ variables ] &gt; ) &lt;- ( &lt; function &gt; -&gt; &lt; pattern &gt; ) ) when &lt; name &gt; ( &lt; [ variables ] &gt; ) = &lt; builder &gt; where pattern and builder reuse the variables . (function -&gt; pattern) is called a view iex&gt; defmodule DoctestTPZ do ...&gt; import PatternMetonyms ...&gt; ...&gt; pattern ( snoc ( x , xs ) &lt;- ( unsnoc ( ) -&gt; { x , xs } ) ) ...&gt; when snoc ( x , xs ) = Enum . reverse ( [ x | Enum . reverse ( xs ) ] ) ...&gt; ...&gt; defp unsnoc ( [ ] ) , do : :error ...&gt; defp unsnoc ( xs ) do ...&gt; [ x | rev_tail ] = Enum . reverse ( xs ) ...&gt; { x , Enum . reverse ( rev_tail ) } ...&gt; end ...&gt; ...&gt; def foo ( xs ) do ...&gt; view xs do ...&gt; snoc ( x , _ ) -&gt; x ...&gt; [ ] -&gt; [ ] ...&gt; end ...&gt; end ...&gt; ...&gt; def bar ( xs ) do ...&gt; view xs do ...&gt; snoc ( x , xs ) -&gt; snoc ( - x , xs ) ...&gt; [ ] -&gt; [ ] ...&gt; end ...&gt; end ...&gt; end iex&gt; DoctestTPZ . foo ( [ 1 , 2 , 3 ] ) 3 iex&gt; DoctestTPZ . bar ( [ 1 , 2 , 3 ] ) [ 1 , 2 , - 3 ] Patterns using a view can not be used with case . Remote function can be used within a view, but the __MODULE__ alias won't work because the expansion is not done at the usage site. It is not yet determined which behavior is desired. iex&gt; defmodule DoctestTPA do ...&gt; import PatternMetonyms ...&gt; ...&gt; pattern rev_head ( x ) &lt;- ( Enum . reverse -&gt; [ x | _ ] ) ...&gt; ...&gt; def blorg ( xs ) do ...&gt; view xs do ...&gt; rev_head ( x ) -&gt; x ...&gt; end ...&gt; end ...&gt; end iex&gt; DoctestTPA . blorg ( [ 1 , 2 , 3 ] ) 3 Unknown yet if anonymous functions can be supported. Guards within a pattern definition is considered undefined behavior, it may work, but it depends on the context. Consider that if the behavior gets a specification, it would be the removal of the possibility of using them. Patterns using a view pattern are the recommend approach. For example: pattern heart ( n ) &lt;- ( less_than_3 -&gt; { :ok , n } ) Patterns can be documented: @doc &quot;&quot;&quot; heart matches when the number is heartfelt &lt;3 &quot;&quot;&quot; pattern heart ( n ) &lt;- ( less_than_3 ( ) -&gt; { :ok , n } ) You can then access the doc as usual: h heart , or h Module.heart .","ref":"PatternMetonyms.html#pattern/1","title":"PatternMetonyms.pattern/1","type":"macro"},{"doc":"Macro substitute for case/2 capable of using pattern metonyms. Custom case able to use pattern metonyms defined with this module. Largely unoptimized, try to avoid side effect in your pattern definitions as using them multiple time in view will repeat them, but might not later on. View pattern ( (function -&gt; pattern) ) may be used raw in here. View patterns are simply a pair of a function associated with a pattern where the function will be applied to the data passed to view and the result will be matched with the pattern. iex&gt; import PatternMetonyms iex&gt; view self ( ) do ...&gt; ( is_pid ( ) -&gt; true ) -&gt; :ok ...&gt; _ -&gt; :ko ...&gt; end :ok Guards can be used outside of the view pattern or the pattern metonym. iex&gt; import PatternMetonyms iex&gt; view - 3 - :rand . uniform ( 2 ) do ...&gt; ( abs ( ) -&gt; x ) when x &gt; 3 -&gt; :ok ...&gt; ( abs ( ) -&gt; x ) when x &lt; 3 -&gt; :ko ...&gt; x -&gt; x ...&gt; end :ok Remote calls can be used directly within a view pattern. iex&gt; import PatternMetonyms iex&gt; view :banana do ...&gt; ( Atom . to_string ( ) -&gt; &quot;ba&quot; &lt;&gt; _ ) -&gt; :ok ...&gt; _ -&gt; :ko ...&gt; end :ok Anonymous functions can be used within a view pattern. They can be either used as stored within a variable: iex&gt; import PatternMetonyms iex&gt; fun = &amp; inspect ( &amp;1 , pretty : &amp;2 ) iex&gt; view :banana do ...&gt; ( fun . ( true ) -&gt; str ) -&gt; str ...&gt; end &quot;:banana&quot; Or defined directly using Kernel.SpecialForms.fn/1 only: iex&gt; import PatternMetonyms iex&gt; view 3 do ...&gt; ( fn x -&gt; x + 2 end -&gt; n ) -&gt; n + 1 ...&gt; end 6","ref":"PatternMetonyms.html#view/2","title":"PatternMetonyms.view/2","type":"macro"},{"doc":"PatternMetonyms Attempt at implementing Pattern Synonyms from https://www.microsoft.com/en-us/research/wp-content/uploads/2016/08/pattern-synonyms-Haskell16.pdf but obviously missing a ton of check, most notably from the type checker. Consider this as a personal project project until 1.0.0, you may open issues, but pull requests will be closed without consideration.","ref":"readme.html","title":"Overview","type":"extras"},{"doc":"This library hides a lot of code, please keep yourself aware that by using it, you are trading a lot of runtime efficiency for a lot of (unmeasured) expressivness power.","ref":"readme.html#warning","title":"Overview - Warning","type":"extras"},{"doc":"The package can be installed by adding pattern_metonyms to your list of dependencies in mix.exs : def deps do [ { :pattern_metonyms , &quot;~&gt; 0.6.1&quot; } ] end The docs can be found at https://hexdocs.pm/pattern_metonyms .","ref":"readme.html#installation","title":"Overview - Installation","type":"extras"},{"doc":"It can be hard to see at first glance why this would be interesting, so I'm going to try to fill example usages over time. Stacks Stacks are easy to implement, and easy to manipulate: just use a list ! You can use it raw: when you want to push something, just cons an element: new_stack = [x | stack] when you want to pop something, just unconse an element: [x | new_stack] = stack We can improve the code by using functions, in order to give names to the operations. This is useful because it gives us a vocabulary, thus when we encounter the code, we don't have to refer to the context, or hold it in our head in order to follow the code: when you want to push something, just call the appropriate function: new_stack = Stack.push(stack, x) when you want to pop something, call the appropriate function and match on the result: {x, new_stack} = Stack.pop(stack) This also has the advantage of functions being first class values, you can pass them around. But has seen, we have to match while knowing what the result of the function will look like. This is fine but it still presents some overhead, both in the execution compared to the raw version, and in the expressivity because of the necessity to match on the tuple. This slight expressivity impediment become more apparent when we want to use pattern matching to control the flow of execution (unless you define more functions to expand the vocabulary of what's possible). Suppose you want to evaluate different expressions depending on if the stack is empty or not: when raw: { x , new_stack } = case stack do [ ] -&gt; { default_value , stack } [ x | new_stack ] -&gt; { x , new_stack } end with functions: { x , new_stack } = case Stack . pop ( stack ) do :empty -&gt; { default_value , stack } t -&gt; t end This is one possibility, but what if Stack.pop/1 was defined more straightforwardly ? You would have to check whether the stack is empty before using it: { x , new_stack } = if Stack . empty? ( stack ) do { default_value , stack } else Stack . pop ( stack ) end Or you could have another version that directly handle the default value, with the potential of introducing some ambiguity if it was defined with a default argument itself. So yes, this is a toy example, so it is absolutly possible to easily define a nice API with functions, but it is possible with macros to define some vocabulary that can mix well with the pattern matching constructs. We this library, you can define a vocabulary, that behaves as the raw code: defmodule Stack do import PatternMetonyms pattern empty = [ ] pattern push ( stack , x ) = [ x | stack ] pattern pop ( x , stack ) = [ x | stack ] end Assuming the module Stack was require'd: when you want to push something: new_stack = Stack.push(stack, x) when you want to pop something: Stack.pop(x, new_stack) = stack when you want to do some control flow: { x , new_stack } = case stack do Stack . empty ( ) -&gt; { default_value , stack } Stack . pop ( x , stack ) -&gt; { x , stack } end That's all fine and dandy, but you don't need to use this library to define these kinds of macro. (Though it isn't as nice looking. :P) So, on to the next part ! Queues Queues aren't as simple as stacks, we don't have a ready made data structure for them we can pattern match on. But we have the erlang module :queue that allows us to construct and manipulate erlang primitives in order to have something that acts like a queue. Like the function section of the stacks, we have a vocabulary available to us to ease the understanding of the code. In the same way, we can't directly use it with pattern matching. But it's not possible to define macros that mix well with the pattern matching construct. Using this library, you can use new control flow constructs, the more interesting one being view/2 . The others building on top of it. Note: Unfortunatly it's currently too hard for me to define an equivalent to the match operator named =/2 . (Mainly because I haven't made the effort.) Because of that, it's difficult to do something as nice as raw code for non conditional execution. Using view patterns, you can directly pattern match with the functions of the module :queue : for control flow: { x , new_queue } = view queue do ( :queue . is_empty ( ) -&gt; true ) -&gt; { default_value , queue } ( :queue . out ( ) -&gt; { { :value , x } , queue } ) -&gt; { x , queue } end Note that :queue.out/1 returns {:empty, queue} in case the queue is empty, allowing to directly pattern match on :queue.out/1 's result: { x , new_queue } = case :queue . out ( queue ) do { :empty , queue } -&gt; { default_value , queue } { { :value , x } , queue } -&gt; { x , queue } end Using patterns definition with view: defmodule Queue do import PatternMetonyms pattern empty &lt;- ( :queue . is_empty ( ) -&gt; true ) pattern pop ( x , queue ) &lt;- ( :queue . out ( ) -&gt; { { :value , x } , queue } ) end for control flow, assuming Queue was require'd: { x , new_queue } = view queue do Queue . empty ( ) -&gt; { default_value , queue } Queue . pop ( x , queue ) -&gt; { x , queue } end This become even more interesting with recursion (though this is still a toy example). Say :queue.to_list/1 doesn't exist, neither :queue.fold/3 , how would you go about defining it ? raw: defmodule Queue do # [...] def to_list ( queue ) do case :queue . out ( queue ) do { :empty , _ } -&gt; [ ] { { :value , x } , queue } -&gt; [ x | to_list ( queue ) end end end with patterns and defv/1 : defmodule Queue do # [...] defv to_list ( empty ( ) ) , do : [ ] defv to_list ( pop ( x , queue ) ) , do : [ x | to_list ( queue ) ] end I hope you are now starting to get an idea about how to play with this lib. More to come when the inspiration does.","ref":"readme.html#examples","title":"Overview - Examples","type":"extras"},{"doc":"Changelog All notable changes to this project will be documented in this file. (Hopefully I won't forget much.)","ref":"changelog.html","title":"Changelog","type":"extras"},{"doc":"","ref":"changelog.html#unreleased","title":"Changelog - Unreleased","type":"extras"},{"doc":"Added Now handling equality patterns.","ref":"changelog.html#0-6-1-2022-05-17","title":"Changelog - 0.6.1 - 2022-05-17","type":"extras"},{"doc":"Added view/2 can now use any function in guards, as long as they return true . Added defv/2 to define named functions capable of the same things as view/2 . Added fnv/1 do defined anonymous functions capable of the same things as view/2 .","ref":"changelog.html#0-6-0-2021-11-03","title":"Changelog - 0.6.0 - 2021-11-03","type":"extras"},{"doc":"Changed Implicitly replace patterns defined using no args with patterns defined using an empty list of args Improved the localisation of the pattern mentioned in the error message for patterns that can only be used inside view/2 Breaking : require parentheses on raw views used with view/2 , and in views used with pattern/1 Reworked view to handle nested pattern usage","ref":"changelog.html#0-5-0-2021-10-20","title":"Changelog - 0.5.0 - 2021-10-20","type":"extras"},{"doc":"Changed Implementation details, using new dependency: Circe Doc updated","ref":"changelog.html#0-4-0-2021-05-14","title":"Changelog - 0.4.0 - 2021-05-14","type":"extras"},{"doc":"Changed Doc updated Added Anonymous functions can now be used with view","ref":"changelog.html#0-3-1-2021-03-11","title":"Changelog - 0.3.1 - 2021-03-11","type":"extras"},{"doc":"Changed Updated this changelog to try the format from keep-a-changelog Reworked view to handle remote calls, updated the doc accordingly","ref":"changelog.html#0-3-0-2020-12-18","title":"Changelog - 0.3.0 - 2020-12-18","type":"extras"},{"doc":"Changed Docs updated Reorganised the order of the patterns definitions, this allows to set the doc attribute for the version used by the programmer, and to hide the internal version used by view Added a vanilla definition for unidirectional pattern using a view, this permit to set a doc attribute usable by the programmer, improve discoverability, and signal with a better error message when the pattern is improperly used (from unknown definition to custom raise) Internal changes Added Guards can now be used with view","ref":"changelog.html#0-2-0-2020-12-18","title":"Changelog - 0.2.0 - 2020-12-18","type":"extras"},{"doc":"Fix view/2 updated to use a unique variable before matching, avoiding repeated function call when it is the value to be matched on pattern/1 : changed ast traversal from pre to post to avoid infinite loops during compilation (was present but forgotten in the changelog) Changed Docs updated Added Changelog creation","ref":"changelog.html#0-1-1-2020-12-04","title":"Changelog - 0.1.1 - 2020-12-04","type":"extras"},{"doc":"Added Base implementation Changelog format inspired by keep-a-changelog","ref":"changelog.html#0-1-0-2020-12-03","title":"Changelog - 0.1.0 - 2020-12-03","type":"extras"}]